# encoding: utf-8
"""
@author:  xingyu liao
@contact: sherlockliao01@gmail.com
"""

from .partial_fc import PartialFC
from .face_baseline import FaceBaseline
from .face_head import FaceHead
from .iresnet import build_iresnet_backbone
