# encoding: utf-8
"""
@author:  xingyu liao
@contact: sherlockliao01@gmail.com
"""

# Attributed datasets
from .pa100k import PA100K
from .market1501attr import Market1501Attr
from .dukemtmcattr import DukeMTMCAttr
