# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['initor_bot_inference',
 'initor_bot_inference.fast_reid',
 'initor_bot_inference.fast_reid.datasets',
 'initor_bot_inference.fast_reid.demo',
 'initor_bot_inference.fast_reid.docs',
 'initor_bot_inference.fast_reid.fastreid',
 'initor_bot_inference.fast_reid.fastreid.config',
 'initor_bot_inference.fast_reid.fastreid.data',
 'initor_bot_inference.fast_reid.fastreid.data.datasets',
 'initor_bot_inference.fast_reid.fastreid.data.samplers',
 'initor_bot_inference.fast_reid.fastreid.data.transforms',
 'initor_bot_inference.fast_reid.fastreid.engine',
 'initor_bot_inference.fast_reid.fastreid.evaluation',
 'initor_bot_inference.fast_reid.fastreid.evaluation.rank_cylib',
 'initor_bot_inference.fast_reid.fastreid.layers',
 'initor_bot_inference.fast_reid.fastreid.modeling',
 'initor_bot_inference.fast_reid.fastreid.modeling.backbones',
 'initor_bot_inference.fast_reid.fastreid.modeling.backbones.regnet',
 'initor_bot_inference.fast_reid.fastreid.modeling.heads',
 'initor_bot_inference.fast_reid.fastreid.modeling.losses',
 'initor_bot_inference.fast_reid.fastreid.modeling.meta_arch',
 'initor_bot_inference.fast_reid.fastreid.solver',
 'initor_bot_inference.fast_reid.fastreid.solver.optim',
 'initor_bot_inference.fast_reid.fastreid.utils',
 'initor_bot_inference.fast_reid.projects.FastAttr',
 'initor_bot_inference.fast_reid.projects.FastAttr.fastattr',
 'initor_bot_inference.fast_reid.projects.FastAttr.fastattr.datasets',
 'initor_bot_inference.fast_reid.projects.FastAttr.fastattr.modeling',
 'initor_bot_inference.fast_reid.projects.FastClas',
 'initor_bot_inference.fast_reid.projects.FastClas.fastclas',
 'initor_bot_inference.fast_reid.projects.FastDistill',
 'initor_bot_inference.fast_reid.projects.FastDistill.fastdistill',
 'initor_bot_inference.fast_reid.projects.FastFace',
 'initor_bot_inference.fast_reid.projects.FastFace.fastface',
 'initor_bot_inference.fast_reid.projects.FastFace.fastface.datasets',
 'initor_bot_inference.fast_reid.projects.FastFace.fastface.modeling',
 'initor_bot_inference.fast_reid.projects.FastRT.pybind_interface',
 'initor_bot_inference.fast_reid.projects.FastRT.tools',
 'initor_bot_inference.fast_reid.projects.FastRetri',
 'initor_bot_inference.fast_reid.projects.FastRetri.fastretri',
 'initor_bot_inference.fast_reid.projects.FastTune',
 'initor_bot_inference.fast_reid.projects.FastTune.autotuner',
 'initor_bot_inference.fast_reid.projects.NAIC20',
 'initor_bot_inference.fast_reid.projects.NAIC20.naic',
 'initor_bot_inference.fast_reid.projects.PartialReID',
 'initor_bot_inference.fast_reid.projects.PartialReID.partialreid',
 'initor_bot_inference.fast_reid.tests',
 'initor_bot_inference.fast_reid.tools',
 'initor_bot_inference.fast_reid.tools.deploy',
 'initor_bot_inference.fast_reid.tools.deploy.Caffe',
 'initor_bot_inference.tools',
 'initor_bot_inference.tracker',
 'initor_bot_inference.tracker.tracking_utils',
 'initor_bot_inference.yolox',
 'initor_bot_inference.yolox.core',
 'initor_bot_inference.yolox.data',
 'initor_bot_inference.yolox.data.datasets',
 'initor_bot_inference.yolox.evaluators',
 'initor_bot_inference.yolox.exp',
 'initor_bot_inference.yolox.exps.default',
 'initor_bot_inference.yolox.exps.example.mot',
 'initor_bot_inference.yolox.layers',
 'initor_bot_inference.yolox.models',
 'initor_bot_inference.yolox.utils']

package_data = \
{'': ['*'],
 'initor_bot_inference.fast_reid': ['.github/*',
                                    '.github/ISSUE_TEMPLATE/*',
                                    '.github/workflows/*',
                                    'configs/*',
                                    'configs/DukeMTMC/*',
                                    'configs/MOT17/*',
                                    'configs/MOT20/*',
                                    'configs/MSMT17/*',
                                    'configs/Market1501/*',
                                    'configs/VERIWild/*',
                                    'configs/VeRi/*',
                                    'configs/VehicleID/*',
                                    'docker/*',
                                    'projects/*',
                                    'projects/CrossDomainReID/*',
                                    'projects/DG-ReID/*',
                                    'projects/FastRT/*',
                                    'projects/FastRT/demo/*',
                                    'projects/FastRT/docker/trt7cu100/*',
                                    'projects/FastRT/docker/trt7cu102/*',
                                    'projects/FastRT/fastrt/*',
                                    'projects/FastRT/fastrt/backbones/*',
                                    'projects/FastRT/fastrt/common/*',
                                    'projects/FastRT/fastrt/engine/*',
                                    'projects/FastRT/fastrt/factory/*',
                                    'projects/FastRT/fastrt/heads/*',
                                    'projects/FastRT/fastrt/layers/*',
                                    'projects/FastRT/fastrt/meta_arch/*',
                                    'projects/FastRT/include/fastrt/*',
                                    'projects/FastRT/third_party/cnpy/*',
                                    'projects/HAA/*'],
 'initor_bot_inference.fast_reid.docs': ['_static/css/*', 'modules/*'],
 'initor_bot_inference.fast_reid.fastreid.modeling.backbones.regnet': ['effnet/*',
                                                                       'regnetx/*',
                                                                       'regnety/*'],
 'initor_bot_inference.fast_reid.projects.FastAttr': ['configs/*'],
 'initor_bot_inference.fast_reid.projects.FastClas': ['configs/*'],
 'initor_bot_inference.fast_reid.projects.FastDistill': ['configs/*'],
 'initor_bot_inference.fast_reid.projects.FastFace': ['configs/*'],
 'initor_bot_inference.fast_reid.projects.FastRT.pybind_interface': ['docker/trt7cu100/*',
                                                                     'docker/trt7cu102_torch160/*'],
 'initor_bot_inference.fast_reid.projects.FastRetri': ['configs/*'],
 'initor_bot_inference.fast_reid.projects.FastTune': ['configs/*'],
 'initor_bot_inference.fast_reid.projects.NAIC20': ['configs/*'],
 'initor_bot_inference.fast_reid.projects.PartialReID': ['configs/*'],
 'initor_bot_inference.fast_reid.tools.deploy': ['test_data/*'],
 'initor_bot_inference.tracker': ['GMC_files/MOT17_ablation/*',
                                  'GMC_files/MOTChallenge/*'],
 'initor_bot_inference.yolox.layers': ['csrc/*', 'csrc/cocoeval/*']}

install_requires = \
['Pillow>=10.0.0,<11.0.0',
 'PyYAML>=6.0,<7.0',
 'easydict>=1.10,<2.0',
 'filterpy>=1.4.5,<2.0.0',
 'gdown>=4.7.1,<5.0.0',
 'h5py>=3.9.0,<4.0.0',
 'lap>=0.4.0,<0.5.0',
 'loguru>=0.7.0,<0.8.0',
 'matplotlib>=3.7.2,<4.0.0',
 'motmetrics>=1.4.0,<2.0.0',
 'ninja>=1.11.1,<2.0.0',
 'numpy==1.23.5',
 'onnx-simplifier==0.3.5',
 'onnx==1.8.1',
 'onnxruntime==1.8.0',
 'opencv-python>=4.8.0,<5.0.0',
 'prettytable>=3.8.0,<4.0.0',
 'scikit-image>=0.21.0,<0.22.0',
 'scikit-learn>=1.3.0,<2.0.0',
 'scipy<1.11.1',
 'tabulate>=0.9.0,<0.10.0',
 'tensorboard>=2.13.0,<3.0.0',
 'termcolor>=2.3.0,<3.0.0',
 'thop>=0.1.1,<0.2.0',
 'torchvision==0.12.0',
 'tqdm>=4.65.0,<5.0.0',
 'yacs>=0.1.8,<0.2.0']

setup_kwargs = {
    'name': 'initor-bot-inference',
    'version': '2.0.0',
    'description': '',
    'long_description': None,
    'author': 's',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
